@@ -210,16 +210,18 @@
 #include "clutter-private.h"
 #include "clutter-debug.h"
 
-#include "json/json-parser.h"
-
 enum
 {
   PROP_0,
 
   PROP_FILENAME_SET,
-  PROP_FILENAME
+  PROP_FILENAME,
+
+  PROP_LAST
 };
 
+static GParamSpec *obj_props[PROP_LAST];
+
 #define CLUTTER_SCRIPT_GET_PRIVATE(obj) \
         (G_TYPE_INSTANCE_GET_PRIVATE ((obj), CLUTTER_TYPE_SCRIPT, ClutterScriptPrivate))
 
@@ -312,26 +314,16 @@ object_info_free (gpointer data)
        * unless we are unmerging in which case we have to destroy
        * the actor to unparent them
        */
-      if (oinfo->object)
+      if (oinfo->object != NULL)
         {
           if (oinfo->is_unmerged)
             {
-              if (oinfo->is_toplevel)
-                g_object_unref (oinfo->object);
-              else
-                {
-                  /* destroy every actor, unless it's the default stage */
-                  if (oinfo->is_stage_default == FALSE &&
-                      CLUTTER_IS_ACTOR (oinfo->object))
-                    clutter_actor_destroy (CLUTTER_ACTOR (oinfo->object));
-                }
-            }
-          else
-            {
-              if (oinfo->is_toplevel)
-                g_object_unref (oinfo->object);
+              if (oinfo->is_actor && !oinfo->is_stage)
+                clutter_actor_destroy (CLUTTER_ACTOR (oinfo->object));
             }
 
+          g_object_unref (oinfo->object);
+
           oinfo->object = NULL;
         }
 
@@ -378,15 +370,11 @@ static void
 clutter_script_class_init (ClutterScriptClass *klass)
 {
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
-  GParamSpec *pspec;
 
   g_type_class_add_private (klass, sizeof (ClutterScriptPrivate));
 
   klass->get_type_from_name = clutter_script_real_get_type_from_name;
 
-  gobject_class->get_property = clutter_script_get_property;
-  gobject_class->finalize = clutter_script_finalize;
-
   /**
    * ClutterScript:filename-set:
    *
@@ -396,12 +384,12 @@ clutter_script_class_init (ClutterScriptClass *klass)
    *
    * Since: 0.6
    */
-  pspec = g_param_spec_boolean ("filename-set",
-                                P_("Filename Set"),
-                                P_("Whether the :filename property is set"),
-                                FALSE,
-                                CLUTTER_PARAM_READABLE);
-  g_object_class_install_property (gobject_class, PROP_FILENAME_SET, pspec);
+  obj_props[PROP_FILENAME_SET] =
+    g_param_spec_boolean ("filename-set",
+                          P_("Filename Set"),
+                          P_("Whether the :filename property is set"),
+                          FALSE,
+                          CLUTTER_PARAM_READABLE);
 
   /**
    * ClutterScript:filename:
@@ -411,12 +399,19 @@ clutter_script_class_init (ClutterScriptClass *klass)
    *
    * Since: 0.6
    */
-  pspec = g_param_spec_string ("filename",
-                               P_("Filename"),
-                               P_("The path of the currently parsed file"),
-                               NULL,
-                               CLUTTER_PARAM_READABLE);
-  g_object_class_install_property (gobject_class, PROP_FILENAME, pspec);
+  obj_props[PROP_FILENAME] =
+    g_param_spec_string ("filename",
+                         P_("Filename"),
+                         P_("The path of the currently parsed file"),
+                         NULL,
+                         CLUTTER_PARAM_READABLE);
+
+  gobject_class->get_property = clutter_script_get_property;
+  gobject_class->finalize = clutter_script_finalize;
+
+  _clutter_object_class_install_properties (gobject_class,
+                                            PROP_LAST,
+                                            obj_props);
 }
 
 static void
@@ -468,7 +463,7 @@ clutter_script_new (void)
  *
  * Return value: on error, zero is returned and @error is set
  *   accordingly. On success, the merge id for the UI definitions is
- *   returned. You can use the merge id with clutter_script_unmerge().
+ *   returned. You can use the merge id with clutter_script_unmerge_objects().
  *
  * Since: 0.6
  */
@@ -517,7 +512,7 @@ clutter_script_load_from_file (ClutterScript  *script,
  *
  * Return value: on error, zero is returned and @error is set
  *   accordingly. On success, the merge id for the UI definitions is
- *   returned. You can use the merge id with clutter_script_unmerge().
+ *   returned. You can use the merge id with clutter_script_unmerge_objects().
  *
  * Since: 0.6
  */
@@ -565,7 +560,7 @@ clutter_script_load_from_data (ClutterScript  *script,
  * Retrieves the object bound to @name. This function does not increment
  * the reference count of the returned object.
  *
- * Return value: : (transfer none): the named object, or %NULL if no object
+ * Return value: (transfer none): the named object, or %NULL if no object
  *   with the given name was available
  *
  * Since: 0.6
@@ -967,7 +962,7 @@ connect_each_object (gpointer key,
 }
 
 /**
- * clutter_script_connect_signals_full:
+ * clutter_script_connect_signals_full: (skip)
  * @script: a #ClutterScript
  * @func: signal connection function
  * @user_data: data to be passed to the signal handlers, or %NULL
@@ -978,7 +973,8 @@ connect_each_object (gpointer key,
  * This function allows to control how the signal handlers are
  * going to be connected to their respective signals. It is meant
  * primarily for language bindings to allow resolving the function
- * names using the native API.
+ * names using the native API, but it can also be used on platforms
+ * that do not support GModule.
  *
  * Applications should use clutter_script_connect_signals().
  *
@@ -1141,9 +1137,10 @@ clutter_script_lookup_filename (ClutterScript *script,
  * Note: this function does not increment the reference count of the
  * objects it returns.
  *
- * Return value: (transfer container) (element-type GObject): a list of #GObject<!-- -->s,
- *   or %NULL. The objects are owned by the #ClutterScript instance. Use g_list_free() on the
- *   returned value when done.
+ * Return value: (transfer container) (element-type GObject.Object): a list
+ *   of #GObject<!-- -->s, or %NULL. The objects are owned by the
+ *   #ClutterScript instance. Use g_list_free() on the returned list when
+ *   done.
  *
  * Since: 0.8.2
  */
