@@ -3,6 +3,7 @@
  *
  * Copyright (C) 2007-2008  Tommi Komulainen <tommi.komulainen@iki.fi>
  * Copyright (C) 2007  OpenedHand Ltd.
+ * Copyright (C) 2011  Crystalnix <vgachkaylo@crystalnix.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -23,11 +24,15 @@
 
 #include "clutter-osx.h"
 #include "clutter-backend-osx.h"
+#include "clutter-device-manager-osx.h"
 #include "clutter-stage-osx.h"
-#include "../clutter-private.h"
+
+#include "clutter-debug.h"
+#include "clutter-private.h"
+#include "clutter-stage-private.h"
+
 #include "cogl/cogl.h"
 
-#include <clutter/clutter-debug.h>
 #import <AppKit/AppKit.h>
 
 G_DEFINE_TYPE (ClutterBackendOSX, clutter_backend_osx, CLUTTER_TYPE_BACKEND)
@@ -43,12 +48,18 @@ clutter_backend_osx_post_parse (ClutterBackend  *backend,
   NSSize size;
   [[prop valueForKey:@"NSDeviceResolution"] getValue:&size];
   CLUTTER_OSX_POOL_RELEASE();
+
   /* setting dpi for backend, it needs by font rendering library */
   if (size.height > 0)
     {
-      clutter_backend_set_resolution (backend, size.height);
+      ClutterSettings *settings = clutter_settings_get_default ();
+      int font_dpi = size.height * 1024;
+
+      g_object_set (settings, "font-dpi", font_dpi, NULL);
+
       return TRUE;
     }
+
   return FALSE;
 }
 
@@ -69,7 +80,7 @@ clutter_backend_osx_create_stage (ClutterBackend  *backend,
 
   CLUTTER_OSX_POOL_ALLOC();
 
-  impl = clutter_stage_osx_new (backend, wrapper);
+  impl = _clutter_stage_osx_new (backend, wrapper);
 
   CLUTTER_NOTE (BACKEND, "create_stage: impl=%p", impl);
 
@@ -78,11 +89,38 @@ clutter_backend_osx_create_stage (ClutterBackend  *backend,
   return impl;
 }
 
+static inline void
+clutter_backend_osx_create_device_manager (ClutterBackendOSX *backend_osx)
+{
+  if (backend_osx->device_manager != NULL)
+    return;
+
+  backend_osx->device_manager = g_object_new (CLUTTER_TYPE_DEVICE_MANAGER_OSX,
+                                              "backend", CLUTTER_BACKEND(backend_osx),
+                                              NULL);
+}
+
+static ClutterDeviceManager *
+clutter_backend_osx_get_device_manager (ClutterBackend *backend)
+{
+  ClutterBackendOSX *backend_osx = CLUTTER_BACKEND_OSX (backend);
+
+  clutter_backend_osx_create_device_manager (backend_osx);
+
+  return backend_osx->device_manager;
+}
+
 static void
 clutter_backend_osx_init_events (ClutterBackend *backend)
 {
+  ClutterBackendOSX *backend_osx = CLUTTER_BACKEND_OSX (backend);
+
+  if (backend_osx->device_manager != NULL)
+    return;
+
   CLUTTER_NOTE (BACKEND, "init_events");
 
+  clutter_backend_osx_create_device_manager (backend_osx);
   _clutter_events_osx_init ();
 }
 
@@ -154,14 +192,25 @@ clutter_backend_osx_ensure_context (ClutterBackend *backend,
 }
 
 static void
-clutter_backend_osx_redraw (ClutterBackend *backend, ClutterStage *wrapper)
+clutter_backend_osx_redraw (ClutterBackend *backend,
+                            ClutterStage   *wrapper)
 {
-  ClutterStageWindow *impl = _clutter_stage_get_window (wrapper);
-  ClutterStageOSX *stage_osx = CLUTTER_STAGE_OSX (impl);
+  ClutterStageWindow *impl;
+  ClutterStageOSX *stage_osx;
+
+  if (CLUTTER_ACTOR_IN_DESTRUCTION (wrapper))
+    return;
+
+  impl = _clutter_stage_get_window (wrapper);
+  if (impl == NULL)
+    return;
+
+  stage_osx = CLUTTER_STAGE_OSX (impl);
 
   CLUTTER_OSX_POOL_ALLOC();
 
-  [stage_osx->view setNeedsDisplay: YES];
+  if (stage_osx->view != NULL)
+    [stage_osx->view setNeedsDisplay: YES];
 
   CLUTTER_OSX_POOL_RELEASE();
 }
@@ -212,13 +261,14 @@ clutter_backend_osx_class_init (ClutterBackendOSXClass *klass)
 
   object_class->dispose = clutter_backend_osx_dispose;
 
-  backend_class->post_parse       = clutter_backend_osx_post_parse;
-  backend_class->get_features     = clutter_backend_osx_get_features;
-  backend_class->create_stage     = clutter_backend_osx_create_stage;
-  backend_class->create_context   = clutter_backend_osx_create_context;
-  backend_class->ensure_context   = clutter_backend_osx_ensure_context;
-  backend_class->init_events      = clutter_backend_osx_init_events;
-  backend_class->redraw           = clutter_backend_osx_redraw;
+  backend_class->post_parse         = clutter_backend_osx_post_parse;
+  backend_class->get_features       = clutter_backend_osx_get_features;
+  backend_class->create_stage       = clutter_backend_osx_create_stage;
+  backend_class->create_context     = clutter_backend_osx_create_context;
+  backend_class->ensure_context     = clutter_backend_osx_ensure_context;
+  backend_class->init_events        = clutter_backend_osx_init_events;
+  backend_class->redraw             = clutter_backend_osx_redraw;
+  backend_class->get_device_manager = clutter_backend_osx_get_device_manager;
 }
 
 GType
