@@ -3,6 +3,7 @@
  *
  * Copyright (C) 2007-2008  Tommi Komulainen <tommi.komulainen@iki.fi>
  * Copyright (C) 2007  OpenedHand Ltd.
+ * Copyright (C) 2011  Crystalnix  <vgachkaylo@gmail.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -135,6 +136,7 @@ clutter_stage_osx_get_wrapper (ClutterStageWindow *stage_window);
 @interface ClutterGLView : NSOpenGLView
 {
   ClutterStageOSX *stage_osx;
+  NSTrackingRectTag tracking_rect;
 }
 - (void) drawRect: (NSRect) bounds;
 @end
@@ -145,6 +147,7 @@ clutter_stage_osx_get_wrapper (ClutterStageWindow *stage_window);
   if ((self = [super initWithFrame:aFrame pixelFormat:aFormat]) != nil)
     {
       self->stage_osx = aStage;
+      tracking_rect = [self addTrackingRect:[self bounds] owner:self userData:NULL assumeInside:NO];
     }
 
   return self;
@@ -176,6 +179,9 @@ clutter_stage_osx_get_wrapper (ClutterStageWindow *stage_window);
   stage_osx->requisition_height = [self bounds].size.height;
   clutter_actor_set_size (CLUTTER_ACTOR (self->stage_osx->wrapper),
                           (int)[self bounds].size.width, (int)[self bounds].size.height);
+
+  [self removeTrackingRect:tracking_rect];
+  tracking_rect = [self addTrackingRect:[self bounds] owner:self userData:NULL assumeInside:NO];
 }
 
 /* Simply forward all events that reach our view to clutter. */
