@@ -19,9 +19,9 @@
  *
  */
 
-#ifdef HAVE_CONFIG_H
 #include "config.h"
-#endif
+
+#include <math.h>
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -33,10 +33,13 @@
 
 #include "clutter-actor-private.h"
 #include "clutter-debug.h"
-#include "clutter-main.h"
-#include "clutter-feature.h"
-#include "clutter-event.h"
+#include "clutter-device-manager-private.h"
 #include "clutter-enum-types.h"
+#include "clutter-event-translator.h"
+#include "clutter-event.h"
+#include "clutter-feature.h"
+#include "clutter-main.h"
+#include "clutter-paint-volume-private.h"
 #include "clutter-private.h"
 #include "clutter-stage-private.h"
 
@@ -48,13 +51,20 @@
 
 #define STAGE_X11_IS_MAPPED(s)  ((((ClutterStageX11 *) (s))->wm_state & STAGE_X11_WITHDRAWN) == 0)
 
-static void clutter_stage_window_iface_init (ClutterStageWindowIface *iface);
+static void clutter_stage_window_iface_init     (ClutterStageWindowIface     *iface);
+static void clutter_event_translator_iface_init (ClutterEventTranslatorIface *iface);
+
+static GHashTable *clutter_stages_by_xid = NULL;
+
+#define clutter_stage_x11_get_type      _clutter_stage_x11_get_type
 
 G_DEFINE_TYPE_WITH_CODE (ClutterStageX11,
                          clutter_stage_x11,
                          G_TYPE_OBJECT,
                          G_IMPLEMENT_INTERFACE (CLUTTER_TYPE_STAGE_WINDOW,
-                                                clutter_stage_window_iface_init));
+                                                clutter_stage_window_iface_init)
+                         G_IMPLEMENT_INTERFACE (CLUTTER_TYPE_EVENT_TRANSLATOR,
+                                                clutter_event_translator_iface_init));
 
 #define _NET_WM_STATE_REMOVE        0    /* remove/unset property */
 #define _NET_WM_STATE_ADD           1    /* add/set property */
@@ -84,9 +94,9 @@ send_wmspec_change_state (ClutterBackendX11 *backend_x11,
   xclient.data.l[4] = 0;
 
   XSendEvent (backend_x11->xdpy, 
-              DefaultRootWindow(backend_x11->xdpy), 
+              DefaultRootWindow (backend_x11->xdpy),
               False,
-              SubstructureRedirectMask|SubstructureNotifyMask,
+              SubstructureRedirectMask | SubstructureNotifyMask,
               (XEvent *)&xclient);
 }
 
@@ -114,24 +124,20 @@ update_state (ClutterStageX11   *stage_x11,
     }
 }
 
-void
+static void
 clutter_stage_x11_fix_window_size (ClutterStageX11 *stage_x11,
                                    gint             new_width,
                                    gint             new_height)
 {
-  ClutterBackend *backend = clutter_get_default_backend ();
-  ClutterBackendX11 *backend_x11;
-  gboolean resize;
-
-  g_return_if_fail (CLUTTER_IS_BACKEND_X11 (backend));
-  backend_x11 = CLUTTER_BACKEND_X11 (backend);
-
-  resize = clutter_stage_get_user_resizable (stage_x11->wrapper);
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
 
   if (stage_x11->xwin != None && !stage_x11->is_foreign_xwin)
     {
       guint min_width, min_height;
       XSizeHints *size_hints;
+      gboolean resize;
+
+      resize = clutter_stage_get_user_resizable (stage_x11->wrapper);
 
       size_hints = XAllocSizeHints();
 
@@ -173,17 +179,13 @@ clutter_stage_x11_fix_window_size (ClutterStageX11 *stage_x11,
     }
 }
 
-void
+static void
 clutter_stage_x11_set_wm_protocols (ClutterStageX11 *stage_x11)
 {
-  ClutterBackend *backend = clutter_get_default_backend ();
-  ClutterBackendX11 *backend_x11;
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
   Atom protocols[2];
   int n = 0;
   
-  g_return_if_fail (CLUTTER_IS_BACKEND_X11 (backend));
-  backend_x11 = CLUTTER_BACKEND_X11 (backend);
-
   protocols[n++] = backend_x11->atom_WM_DELETE_WINDOW;
   protocols[n++] = backend_x11->atom_NET_WM_PING;
 
@@ -194,12 +196,8 @@ static void
 clutter_stage_x11_get_geometry (ClutterStageWindow *stage_window,
                                 ClutterGeometry    *geometry)
 {
-  ClutterBackend *backend = clutter_get_default_backend ();
-  ClutterBackendX11 *backend_x11;
   ClutterStageX11 *stage_x11 = CLUTTER_STAGE_X11 (stage_window);
-
-  g_return_if_fail (CLUTTER_IS_BACKEND_X11 (backend));
-  backend_x11 = CLUTTER_BACKEND_X11 (backend);
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
 
   /* If we're fullscreen, return the size of the display. */
   if ((stage_x11->state & CLUTTER_STAGE_STATE_FULLSCREEN) &&
@@ -220,9 +218,8 @@ clutter_stage_x11_resize (ClutterStageWindow *stage_window,
                           gint                width,
                           gint                height)
 {
-  ClutterBackend *backend = clutter_get_default_backend ();
-  ClutterBackendX11 *backend_x11;
   ClutterStageX11 *stage_x11 = CLUTTER_STAGE_X11 (stage_window);
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
   gboolean resize;
 
   if (stage_x11->is_foreign_xwin)
@@ -243,9 +240,6 @@ clutter_stage_x11_resize (ClutterStageWindow *stage_window,
 
   resize = clutter_stage_get_user_resizable (stage_x11->wrapper);
 
-  g_return_if_fail (CLUTTER_IS_BACKEND_X11 (backend));
-  backend_x11 = CLUTTER_BACKEND_X11 (backend);
-
   if (width == 0 || height == 0)
     {
       /* Should not happen, if this turns up we need to debug it and
@@ -289,8 +283,7 @@ clutter_stage_x11_resize (ClutterStageWindow *stage_window,
 static inline void
 set_wm_pid (ClutterStageX11 *stage_x11)
 {
-  ClutterBackend *backend = clutter_get_default_backend ();
-  ClutterBackendX11 *backend_x11 = CLUTTER_BACKEND_X11 (backend);
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
   long pid;
 
   if (stage_x11->xwin == None || stage_x11->is_foreign_xwin)
@@ -314,11 +307,7 @@ set_wm_pid (ClutterStageX11 *stage_x11)
 static inline void
 set_wm_title (ClutterStageX11 *stage_x11)
 {
-  ClutterBackend *backend = clutter_get_default_backend ();
-  ClutterBackendX11 *backend_x11;
-
-  g_return_if_fail (CLUTTER_IS_BACKEND_X11 (backend));
-  backend_x11 = CLUTTER_BACKEND_X11 (backend);
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
 
   if (stage_x11->xwin == None || stage_x11->is_foreign_xwin)
     return;
@@ -345,11 +334,7 @@ set_wm_title (ClutterStageX11 *stage_x11)
 static inline void
 set_cursor_visible (ClutterStageX11 *stage_x11)
 {
-  ClutterBackend *backend = clutter_get_default_backend ();
-  ClutterBackendX11 *backend_x11;
-
-  g_return_if_fail (CLUTTER_IS_BACKEND_X11 (backend));
-  backend_x11 = CLUTTER_BACKEND_X11 (backend);
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
 
   if (stage_x11->xwin == None)
     return;
@@ -393,11 +378,68 @@ static gboolean
 clutter_stage_x11_realize (ClutterStageWindow *stage_window)
 {
   ClutterStageX11 *stage_x11 = CLUTTER_STAGE_X11 (stage_window);
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
+  ClutterDeviceManager *device_manager;
+  int event_flags;
 
   set_wm_pid (stage_x11);
   set_wm_title (stage_x11);
   set_cursor_visible (stage_x11);
 
+
+  /* the masks for the events we want to select on a stage window;
+   * KeyPressMask and KeyReleaseMask are necessary even with XI1
+   * because key events are broken with that extension, and will
+   * be fixed by XI2
+   */
+  event_flags = StructureNotifyMask
+              | FocusChangeMask
+              | ExposureMask
+              | PropertyChangeMask
+              | EnterWindowMask
+              | LeaveWindowMask
+              | KeyPressMask
+              | KeyReleaseMask
+              | ButtonPressMask
+              | ButtonReleaseMask
+              | PointerMotionMask;
+
+  /* we unconditionally select input events even with event retrieval
+   * disabled because we need to guarantee that the Clutter internal
+   * state is maintained when calling clutter_x11_handle_event() without
+   * requiring applications or embedding toolkits to select events
+   * themselves. if we did that, we'd have to document the events to be
+   * selected, and also update applications and embedding toolkits each
+   * time we added a new mask, or a new class of events.
+   *
+   * see: http://bugzilla.clutter-project.org/show_bug.cgi?id=998
+   * for the rationale of why we did conditional selection. it is now
+   * clear that a compositor should clear out the input region, since
+   * it cannot assume a perfectly clean slate coming from us.
+   *
+   * see: http://bugzilla.clutter-project.org/show_bug.cgi?id=2228
+   * for an example of things that break if we do conditional event
+   * selection.
+   */
+  XSelectInput (backend_x11->xdpy, stage_x11->xwin, event_flags);
+
+  /* input events also depent on the actual device, so we need to
+   * use the device manager to let every device select them, using
+   * the event mask we passed to XSelectInput as the template
+   */
+  device_manager = clutter_device_manager_get_default ();
+  _clutter_device_manager_select_stage_events (device_manager,
+                                               stage_x11->wrapper,
+                                               event_flags);
+
+  /* no user resize.. */
+  clutter_stage_x11_fix_window_size (stage_x11,
+                                     stage_x11->xwin_width,
+                                     stage_x11->xwin_height);
+  clutter_stage_x11_set_wm_protocols (stage_x11);
+
+  CLUTTER_NOTE (BACKEND, "Successfully realized stage");
+
   return TRUE;
 }
 
@@ -405,18 +447,18 @@ static void
 clutter_stage_x11_set_fullscreen (ClutterStageWindow *stage_window,
                                   gboolean            is_fullscreen)
 {
-  ClutterBackend *backend = clutter_get_default_backend ();
-  ClutterBackendX11 *backend_x11;
   ClutterStageX11 *stage_x11 = CLUTTER_STAGE_X11 (stage_window);
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
   ClutterStage *stage = stage_x11->wrapper;
+  gboolean was_fullscreen;
 
-  g_return_if_fail (CLUTTER_IS_BACKEND_X11 (backend));
-  backend_x11 = CLUTTER_BACKEND_X11 (backend);
-
-  if (stage == NULL)
+  if (stage == NULL || CLUTTER_ACTOR_IN_DESTRUCTION (stage))
     return;
 
-  if (!!(stage_x11->state & CLUTTER_STAGE_STATE_FULLSCREEN) == is_fullscreen)
+  was_fullscreen = ((stage_x11->state & CLUTTER_STAGE_STATE_FULLSCREEN) != 0);
+  is_fullscreen = !!is_fullscreen;
+
+  if (was_fullscreen == is_fullscreen)
     return;
 
   CLUTTER_NOTE (BACKEND, "%ssetting fullscreen", is_fullscreen ? "" : "un");
@@ -511,7 +553,7 @@ clutter_stage_x11_set_cursor_visible (ClutterStageWindow *stage_window,
 {
   ClutterStageX11 *stage_x11 = CLUTTER_STAGE_X11 (stage_window);
 
-  stage_x11->is_cursor_visible = (cursor_visible == TRUE);
+  stage_x11->is_cursor_visible = !!cursor_visible;
   set_cursor_visible (stage_x11);
 }
 
@@ -540,8 +582,7 @@ clutter_stage_x11_set_user_resizable (ClutterStageWindow *stage_window,
 static inline void
 update_wm_hints (ClutterStageX11 *stage_x11)
 {
-  ClutterBackend *backend;
-  ClutterBackendX11 *backend_x11;
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
   XWMHints wm_hints;
 
   if (stage_x11->wm_state & STAGE_X11_WITHDRAWN)
@@ -550,10 +591,6 @@ update_wm_hints (ClutterStageX11 *stage_x11)
   if (stage_x11->is_foreign_xwin)
     return;
 
-  backend = clutter_get_default_backend ();
-  g_assert (CLUTTER_IS_BACKEND_X11 (backend));
-  backend_x11 = CLUTTER_BACKEND_X11 (backend);
-
   wm_hints.flags = StateHint | InputHint;
   wm_hints.initial_state = NormalState;
   wm_hints.input = stage_x11->accept_focus ? True : False;
@@ -594,12 +631,8 @@ static void
 clutter_stage_x11_show (ClutterStageWindow *stage_window,
                         gboolean            do_raise)
 {
-  ClutterBackend *backend = clutter_get_default_backend ();
-  ClutterBackendX11 *backend_x11;
   ClutterStageX11 *stage_x11 = CLUTTER_STAGE_X11 (stage_window);
-
-  g_return_if_fail (CLUTTER_IS_BACKEND_X11 (backend));
-  backend_x11 = CLUTTER_BACKEND_X11 (backend);
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
 
   if (stage_x11->xwin != None)
     {
@@ -637,12 +670,8 @@ clutter_stage_x11_show (ClutterStageWindow *stage_window,
 static void
 clutter_stage_x11_hide (ClutterStageWindow *stage_window)
 {
-  ClutterBackend *backend = clutter_get_default_backend ();
-  ClutterBackendX11 *backend_x11;
   ClutterStageX11 *stage_x11 = CLUTTER_STAGE_X11 (stage_window);
-
-  g_return_if_fail (CLUTTER_IS_BACKEND_X11 (backend));
-  backend_x11 = CLUTTER_BACKEND_X11 (backend);
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
 
   if (stage_x11->xwin != None)
     {
@@ -677,6 +706,11 @@ clutter_stage_x11_finalize (GObject *gobject)
 static void
 clutter_stage_x11_dispose (GObject *gobject)
 {
+  ClutterEventTranslator *translator = CLUTTER_EVENT_TRANSLATOR (gobject);
+  ClutterBackendX11 *backend_x11 = CLUTTER_STAGE_X11 (gobject)->backend;
+
+  _clutter_backend_x11_remove_event_translator (backend_x11, translator);
+
   G_OBJECT_CLASS (clutter_stage_x11_parent_class)->dispose (gobject);
 }
 
@@ -724,6 +758,344 @@ clutter_stage_window_iface_init (ClutterStageWindowIface *iface)
   iface->realize = clutter_stage_x11_realize;
 }
 
+static inline void
+set_user_time (ClutterBackendX11 *backend_x11,
+               ClutterStageX11   *stage_x11,
+               long               timestamp)
+{
+  if (timestamp != CLUTTER_CURRENT_TIME)
+    {
+      XChangeProperty (backend_x11->xdpy,
+                       stage_x11->xwin,
+                       backend_x11->atom_NET_WM_USER_TIME,
+                       XA_CARDINAL, 32,
+                       PropModeReplace,
+                       (unsigned char *) &timestamp, 1);
+    }
+}
+
+static gboolean
+handle_wm_protocols_event (ClutterBackendX11 *backend_x11,
+                           ClutterStageX11   *stage_x11,
+                           XEvent            *xevent)
+{
+  Atom atom = (Atom) xevent->xclient.data.l[0];
+
+  if (atom == backend_x11->atom_WM_DELETE_WINDOW &&
+      xevent->xany.window == stage_x11->xwin)
+    {
+      /* the WM_DELETE_WINDOW is a request: we do not destroy
+       * the window right away, as it might contain vital data;
+       * we relay the event to the application and we let it
+       * handle the request
+       */
+      CLUTTER_NOTE (EVENT, "Delete stage %s[%p], win:0x%x",
+                    _clutter_actor_get_debug_name (CLUTTER_ACTOR (stage_x11->wrapper)),
+                    stage_x11->wrapper,
+                    (unsigned int) stage_x11->xwin);
+
+      set_user_time (backend_x11, stage_x11, xevent->xclient.data.l[1]);
+
+      return TRUE;
+    }
+  else if (atom == backend_x11->atom_NET_WM_PING &&
+           xevent->xany.window == stage_x11->xwin)
+    {
+      XClientMessageEvent xclient = xevent->xclient;
+
+      xclient.window = backend_x11->xwin_root;
+      XSendEvent (backend_x11->xdpy, xclient.window,
+                  False,
+                  SubstructureRedirectMask | SubstructureNotifyMask,
+                  (XEvent *) &xclient);
+      return FALSE;
+    }
+
+  /* do not send any of the WM_PROTOCOLS events to the queue */
+  return FALSE;
+}
+
+static gboolean
+clipped_redraws_cool_off_cb (void *data)
+{
+  ClutterStageX11 *stage_x11 = data;
+
+  stage_x11->clipped_redraws_cool_off = 0;
+
+  return FALSE;
+}
+
+static ClutterTranslateReturn
+clutter_stage_x11_translate_event (ClutterEventTranslator *translator,
+                                   gpointer                native,
+                                   ClutterEvent           *event)
+{
+  ClutterStageX11 *stage_x11 = CLUTTER_STAGE_X11 (translator);
+  ClutterTranslateReturn res = CLUTTER_TRANSLATE_CONTINUE;
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
+  Window stage_xwindow = stage_x11->xwin;
+  XEvent *xevent = native;
+  ClutterStage *stage;
+
+  stage = clutter_x11_get_stage_from_window (xevent->xany.window);
+  if (stage == NULL)
+    return CLUTTER_TRANSLATE_CONTINUE;
+
+  switch (xevent->type)
+    {
+    case ConfigureNotify:
+      if (!stage_x11->is_foreign_xwin)
+        {
+          gboolean size_changed = FALSE;
+
+          CLUTTER_NOTE (BACKEND, "ConfigureNotify[%x] (%d, %d)",
+                        (unsigned int) stage_x11->xwin,
+                        xevent->xconfigure.width,
+                        xevent->xconfigure.height);
+
+          /* When fullscreen, we'll keep the xwin_width/height
+             variables to track the old size of the window and we'll
+             assume all ConfigureNotifies constitute a size change */
+          if ((stage_x11->state & CLUTTER_STAGE_STATE_FULLSCREEN))
+            size_changed = TRUE;
+          else if ((stage_x11->xwin_width != xevent->xconfigure.width) ||
+                   (stage_x11->xwin_height != xevent->xconfigure.height))
+            {
+              size_changed = TRUE;
+              stage_x11->xwin_width = xevent->xconfigure.width;
+              stage_x11->xwin_height = xevent->xconfigure.height;
+            }
+
+          clutter_actor_set_size (CLUTTER_ACTOR (stage),
+                                  xevent->xconfigure.width,
+                                  xevent->xconfigure.height);
+
+          CLUTTER_UNSET_PRIVATE_FLAGS (stage_x11->wrapper, CLUTTER_IN_RESIZE);
+
+          if (size_changed)
+            {
+              /* XXX: This is a workaround for a race condition when
+               * resizing windows while there are in-flight
+               * glXCopySubBuffer blits happening.
+               *
+               * The problem stems from the fact that rectangles for the
+               * blits are described relative to the bottom left of the
+               * window and because we can't guarantee control over the X
+               * window gravity used when resizing so the gravity is
+               * typically NorthWest not SouthWest.
+               *
+               * This means if you grow a window vertically the server
+               * will make sure to place the old contents of the window
+               * at the top-left/north-west of your new larger window, but
+               * that may happen asynchronous to GLX preparing to do a
+               * blit specified relative to the bottom-left/south-west of
+               * the window (based on the old smaller window geometry).
+               *
+               * When the GLX issued blit finally happens relative to the
+               * new bottom of your window, the destination will have
+               * shifted relative to the top-left where all the pixels you
+               * care about are so it will result in a nasty artefact
+               * making resizing look very ugly!
+               *
+               * We can't currently fix this completely, in-part because
+               * the window manager tends to trample any gravity we might
+               * set.  This workaround instead simply disables blits for a
+               * while if we are notified of any resizes happening so if
+               * the user is resizing a window via the window manager then
+               * they may see an artefact for one frame but then we will
+               * fallback to redrawing the full stage until the cooling
+               * off period is over.
+               */
+              if (stage_x11->clipped_redraws_cool_off)
+                g_source_remove (stage_x11->clipped_redraws_cool_off);
+
+              stage_x11->clipped_redraws_cool_off =
+                g_timeout_add_seconds (1, clipped_redraws_cool_off_cb,
+                                       stage_x11);
+
+              /* Queue a relayout - we want glViewport to be called
+               * with the correct values, and this is done in ClutterStage
+               * via _cogl_onscreen_clutter_backend_set_size ().
+               *
+               * We queue a relayout, because if this ConfigureNotify is
+               * in response to a size we set in the application, the
+               * set_size() call above is essentially a null-op.
+               *
+               * Make sure we do this only when the size has changed,
+               * otherwise we end up relayouting on window moves.
+               */
+              clutter_actor_queue_relayout (CLUTTER_ACTOR (stage));
+
+              /* the resize process is complete, so we can ask the stage
+               * to set up the GL viewport with the new size
+               */
+              clutter_stage_ensure_viewport (stage);
+            }
+        }
+      break;
+
+    case PropertyNotify:
+      if (xevent->xproperty.atom == backend_x11->atom_NET_WM_STATE &&
+          xevent->xproperty.window == stage_xwindow &&
+          !stage_x11->is_foreign_xwin)
+        {
+          Atom     type;
+          gint     format;
+          gulong   n_items, bytes_after;
+          guchar  *data = NULL;
+          gboolean fullscreen_set = FALSE;
+
+          clutter_x11_trap_x_errors ();
+          XGetWindowProperty (backend_x11->xdpy, stage_xwindow,
+                              backend_x11->atom_NET_WM_STATE,
+                              0, G_MAXLONG,
+                              False, XA_ATOM,
+                              &type, &format, &n_items,
+                              &bytes_after, &data);
+          clutter_x11_untrap_x_errors ();
+
+          if (type != None && data != NULL)
+            {
+              Atom *atoms = (Atom *) data;
+              gulong i;
+              gboolean is_fullscreen = FALSE;
+
+              for (i = 0; i < n_items; i++)
+                {
+                  if (atoms[i] == backend_x11->atom_NET_WM_STATE_FULLSCREEN)
+                    fullscreen_set = TRUE;
+                }
+
+              is_fullscreen =
+                (stage_x11->state & CLUTTER_STAGE_STATE_FULLSCREEN);
+
+              if (fullscreen_set != is_fullscreen)
+                {
+                  if (fullscreen_set)
+                    stage_x11->state |= CLUTTER_STAGE_STATE_FULLSCREEN;
+                  else
+                    stage_x11->state &= ~CLUTTER_STAGE_STATE_FULLSCREEN;
+
+                  stage_x11->fullscreening = fullscreen_set;
+
+                  event->any.type = CLUTTER_STAGE_STATE;
+                  event->any.source = CLUTTER_ACTOR (stage);
+                  event->any.stage = stage;
+                  event->stage_state.changed_mask =
+                    CLUTTER_STAGE_STATE_FULLSCREEN;
+                  event->stage_state.new_state = stage_x11->state;
+
+                  res = CLUTTER_TRANSLATE_QUEUE;
+                }
+
+              XFree (data);
+            }
+        }
+      break;
+
+    case FocusIn:
+      if (!(stage_x11->state & CLUTTER_STAGE_STATE_ACTIVATED))
+        {
+          /* TODO: check the detail? */
+          stage_x11->state |= CLUTTER_STAGE_STATE_ACTIVATED;
+
+          event->type = CLUTTER_STAGE_STATE;
+          event->any.source = CLUTTER_ACTOR (stage);
+          event->any.stage = stage;
+          event->stage_state.changed_mask = CLUTTER_STAGE_STATE_ACTIVATED;
+          event->stage_state.new_state = stage_x11->state;
+
+          res = CLUTTER_TRANSLATE_QUEUE;
+        }
+      break;
+
+    case FocusOut:
+      if (stage_x11->state & CLUTTER_STAGE_STATE_ACTIVATED)
+        {
+          /* TODO: check the detail? */
+          stage_x11->state &= ~CLUTTER_STAGE_STATE_ACTIVATED;
+
+          event->any.type = CLUTTER_STAGE_STATE;
+          event->any.source = CLUTTER_ACTOR (stage);
+          event->any.stage = stage;
+          event->stage_state.changed_mask = CLUTTER_STAGE_STATE_ACTIVATED;
+          event->stage_state.new_state = stage_x11->state;
+
+          res = CLUTTER_TRANSLATE_QUEUE;
+        }
+      break;
+
+    case Expose:
+      {
+        XExposeEvent *expose = (XExposeEvent *) xevent;
+        ClutterPaintVolume clip;
+        ClutterVertex origin;
+
+        CLUTTER_NOTE (EVENT,
+                      "expose for stage: %s[%p], win:0x%x - "
+                      "redrawing area (x: %d, y: %d, width: %d, height: %d)",
+                      _clutter_actor_get_debug_name (CLUTTER_ACTOR (stage)),
+                      stage,
+                      (unsigned int) stage_xwindow,
+                      expose->x,
+                      expose->y,
+                      expose->width,
+                      expose->height);
+
+        origin.x = expose->x;
+        origin.y = expose->y;
+        origin.z = 0;
+
+        _clutter_paint_volume_init_static (CLUTTER_ACTOR (stage), &clip);
+
+        clutter_paint_volume_set_origin (&clip, &origin);
+        clutter_paint_volume_set_width (&clip, expose->width);
+        clutter_paint_volume_set_height (&clip, expose->height);
+
+        _clutter_actor_queue_redraw_with_clip (CLUTTER_ACTOR (stage), 0, &clip);
+
+        clutter_paint_volume_free (&clip);
+      }
+      break;
+
+    case DestroyNotify:
+      CLUTTER_NOTE (EVENT,
+                    "Destroy notification received for stage %s[%p], win:0x%x",
+                    _clutter_actor_get_debug_name (CLUTTER_ACTOR (stage)),
+                    stage,
+                    (unsigned int) stage_xwindow);
+      event->any.type = CLUTTER_DESTROY_NOTIFY;
+      event->any.stage = stage;
+      res = CLUTTER_TRANSLATE_QUEUE;
+      break;
+
+    case ClientMessage:
+      CLUTTER_NOTE (EVENT, "Client message for stage %s[%p], win:0x%x",
+                    _clutter_actor_get_debug_name (CLUTTER_ACTOR (stage)),
+                    stage,
+                    (unsigned int) stage_xwindow);
+      if (handle_wm_protocols_event (backend_x11, stage_x11, xevent))
+        {
+          event->any.type = CLUTTER_DELETE;
+          event->any.stage = stage;
+          res = CLUTTER_TRANSLATE_QUEUE;
+        }
+      break;
+
+    default:
+      res = CLUTTER_TRANSLATE_CONTINUE;
+      break;
+    }
+
+  return res;
+}
+
+static void
+clutter_event_translator_iface_init (ClutterEventTranslatorIface *iface)
+{
+  iface->translate_event = clutter_stage_x11_translate_event;
+}
+
 /**
  * clutter_x11_get_stage_window:
  * @stage: a #ClutterStage
@@ -761,24 +1133,16 @@ clutter_x11_get_stage_window (ClutterStage *stage)
 ClutterStage *
 clutter_x11_get_stage_from_window (Window win)
 {
-  ClutterStageManager *stage_manager;
-  const GSList        *stages, *s;
+  ClutterStageX11 *stage_x11;
 
-  stage_manager = clutter_stage_manager_get_default ();
-  stages = clutter_stage_manager_peek_stages (stage_manager);
+  if (clutter_stages_by_xid == NULL)
+    return NULL;
 
-  /* XXX: might use a hash here for performance resaon */
-  for (s = stages; s != NULL; s = s->next)
-    {
-      ClutterStage *stage = s->data;
-      ClutterStageWindow *impl;
+  stage_x11 = g_hash_table_lookup (clutter_stages_by_xid,
+                                   GINT_TO_POINTER (win));
 
-      impl = _clutter_stage_get_window (stage);
-      g_assert (CLUTTER_IS_STAGE_X11 (impl));
-
-      if (CLUTTER_STAGE_X11 (impl)->xwin == win)
-        return stage;
-    }
+  if (stage_x11 != NULL)
+    return stage_x11->wrapper;
 
   return NULL;
 }
@@ -811,7 +1175,7 @@ clutter_x11_get_stage_visual (ClutterStage *stage)
   g_return_val_if_fail (CLUTTER_IS_BACKEND_X11 (backend), NULL);
   backend_x11 = CLUTTER_BACKEND_X11 (backend);
 
-  return clutter_backend_x11_get_visual_info (backend_x11);
+  return _clutter_backend_x11_get_visual_info (backend_x11);
 }
 
 typedef struct {
@@ -826,8 +1190,7 @@ set_foreign_window_callback (ClutterActor *actor,
                              void         *data)
 {
   ForeignWindowData *fwd = data;
-  ClutterBackend *backend = clutter_get_default_backend ();
-  ClutterBackendX11 *backend_x11 = CLUTTER_BACKEND_X11 (backend);
+  ClutterBackendX11 *backend_x11 = fwd->stage_x11->backend;
 
   CLUTTER_NOTE (BACKEND, "Setting foreign window (0x%x)",
                 (unsigned int) fwd->xwindow);
@@ -847,6 +1210,13 @@ set_foreign_window_callback (ClutterActor *actor,
 
   clutter_actor_set_geometry (actor, &fwd->geom);
 
+  if (clutter_stages_by_xid == NULL)
+    clutter_stages_by_xid = g_hash_table_new (NULL, NULL);
+
+  g_hash_table_insert (clutter_stages_by_xid,
+                       GINT_TO_POINTER (fwd->stage_x11->xwin),
+                       fwd->stage_x11);
+
   /* calling this with the stage unrealized will unset the stage
    * from the GL context; once the stage is realized the GL context
    * will be set again
@@ -869,7 +1239,6 @@ gboolean
 clutter_x11_set_stage_foreign (ClutterStage *stage,
                                Window        xwindow)
 {
-  ClutterBackend *backend = clutter_get_default_backend ();
   ClutterBackendX11 *backend_x11;
   ClutterStageX11 *stage_x11;
   ClutterStageWindow *impl;
@@ -881,16 +1250,15 @@ clutter_x11_set_stage_foreign (ClutterStage *stage,
   ForeignWindowData fwd;
   XVisualInfo *xvisinfo;
 
-  g_return_val_if_fail (CLUTTER_IS_BACKEND_X11 (backend), FALSE);
-  backend_x11 = CLUTTER_BACKEND_X11 (backend);
-
   g_return_val_if_fail (CLUTTER_IS_STAGE (stage), FALSE);
   g_return_val_if_fail (!CLUTTER_ACTOR_IN_DESTRUCTION (stage), FALSE);
   g_return_val_if_fail (xwindow != None, FALSE);
 
-  actor = CLUTTER_ACTOR (stage);
+  impl = _clutter_stage_get_window (stage);
+  stage_x11 = CLUTTER_STAGE_X11 (impl);
+  backend_x11 = stage_x11->backend;
 
-  xvisinfo = clutter_backend_x11_get_visual_info (backend_x11);
+  xvisinfo = _clutter_backend_x11_get_visual_info (backend_x11);
   g_return_val_if_fail (xvisinfo != NULL, FALSE);
 
   clutter_x11_trap_x_errors ();
@@ -925,9 +1293,6 @@ clutter_x11_set_stage_foreign (ClutterStage *stage,
       return FALSE;
     }
 
-  impl = _clutter_stage_get_window (stage);
-  stage_x11 = CLUTTER_STAGE_X11 (impl);
-
   fwd.stage_x11 = stage_x11;
   fwd.xwindow = xwindow;
 
@@ -942,6 +1307,8 @@ clutter_x11_set_stage_foreign (ClutterStage *stage,
   fwd.geom.width = width;
   fwd.geom.height = height;
 
+  actor = CLUTTER_ACTOR (stage);
+
   _clutter_actor_rerealize (actor,
                             set_foreign_window_callback,
                             &fwd);
@@ -955,7 +1322,145 @@ clutter_x11_set_stage_foreign (ClutterStage *stage,
    * in the Cogl viewport changing when _clutter_do_redraw
    * calls _clutter_stage_maybe_setup_viewport().
    */
-  clutter_actor_queue_relayout (CLUTTER_ACTOR (stage));
+  clutter_actor_queue_relayout (actor);
 
   return TRUE;
 }
+
+void
+_clutter_stage_x11_destroy_window_untrapped (ClutterStageX11 *stage_x11)
+{
+  Window xwin = stage_x11->xwin;
+
+  if (clutter_stages_by_xid != NULL)
+    g_hash_table_remove (clutter_stages_by_xid, GINT_TO_POINTER (xwin));
+
+  if (!stage_x11->is_foreign_xwin && xwin != None)
+    {
+      ClutterBackendX11 *backend_x11 = stage_x11->backend;
+
+      g_assert (clutter_stages_by_xid != NULL);
+
+      XDestroyWindow (backend_x11->xdpy, xwin);
+      stage_x11->xwin = None;
+    }
+  else
+    stage_x11->xwin = None;
+}
+
+void
+_clutter_stage_x11_destroy_window (ClutterStageX11 *stage_x11)
+{
+  if (stage_x11->xwin == None)
+    return;
+
+  clutter_x11_trap_x_errors ();
+
+  _clutter_stage_x11_destroy_window_untrapped (stage_x11);
+
+  clutter_x11_untrap_x_errors ();
+}
+
+gboolean
+_clutter_stage_x11_create_window (ClutterStageX11 *stage_x11)
+{
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
+  XSetWindowAttributes xattr;
+  XVisualInfo *xvisinfo;
+  unsigned long mask;
+  gfloat width, height;
+
+  if (stage_x11->xwin != None)
+    return TRUE;
+
+  CLUTTER_NOTE (MISC, "Creating stage X window");
+
+  xvisinfo = _clutter_backend_x11_get_visual_info (backend_x11);
+  if (xvisinfo == NULL)
+    {
+      g_critical ("Unable to find suitable GL visual.");
+      return FALSE;
+    }
+
+  /* window attributes */
+  xattr.background_pixel = WhitePixel (backend_x11->xdpy,
+                                       backend_x11->xscreen_num);
+  xattr.border_pixel = 0;
+  xattr.colormap = XCreateColormap (backend_x11->xdpy,
+                                    backend_x11->xwin_root,
+                                    xvisinfo->visual,
+                                    AllocNone);
+  mask = CWBorderPixel | CWColormap;
+
+  /* Call get_size - this will either get the geometry size (which
+   * before we create the window is set to 640x480), or if a size
+   * is set, it will get that. This lets you set a size on the
+   * stage before it's realized.
+   *
+   * we also round to the nearest integer because stage sizes
+   * should always be in pixels
+   */
+  clutter_actor_get_size (CLUTTER_ACTOR (stage_x11->wrapper), &width, &height);
+  stage_x11->xwin_width = floorf (width + 0.5);
+  stage_x11->xwin_height = floorf (height + 0.5);
+
+  stage_x11->xwin = XCreateWindow (backend_x11->xdpy,
+                                   backend_x11->xwin_root,
+                                   0, 0,
+                                   stage_x11->xwin_width,
+                                   stage_x11->xwin_height,
+                                   0,
+                                   xvisinfo->depth,
+                                   InputOutput,
+                                   xvisinfo->visual,
+                                   mask, &xattr);
+
+  CLUTTER_NOTE (BACKEND, "Stage [%p], window: 0x%x, size: %dx%d",
+                stage_x11,
+                (unsigned int) stage_x11->xwin,
+                stage_x11->xwin_width,
+                stage_x11->xwin_height);
+
+  XFree (xvisinfo);
+
+  if (clutter_stages_by_xid == NULL)
+    clutter_stages_by_xid = g_hash_table_new (NULL, NULL);
+
+  g_hash_table_insert (clutter_stages_by_xid,
+                       GINT_TO_POINTER (stage_x11->xwin),
+                       stage_x11);
+
+  return TRUE;
+}
+
+void
+_clutter_stage_x11_set_user_time (ClutterStageX11 *stage_x11,
+                                  guint32          user_time)
+{
+  set_user_time (stage_x11->backend, stage_x11, user_time);
+}
+
+gboolean
+_clutter_stage_x11_get_root_coords (ClutterStageX11 *stage_x11,
+                                    gint            *root_x,
+                                    gint            *root_y)
+{
+  ClutterBackendX11 *backend_x11 = stage_x11->backend;
+  gint return_val;
+  Window child;
+  gint tx, ty;
+
+  return_val = XTranslateCoordinates (backend_x11->xdpy,
+                                      stage_x11->xwin,
+                                      backend_x11->xwin_root,
+                                      0, 0, &tx, &ty,
+                                      &child);
+
+  if (root_x)
+    *root_x = tx;
+
+  if (root_y)
+    *root_y = ty;
+
+  return (return_val == 0);
+}
